Require Import
        Coq.Sets.Ensembles
        Coq.Lists.List
        Coq.Strings.String
        SfLib
        ADT.ADTSig
        ADT.ADT
        ADT.BoundedLookup
        ADT.ilist
        ADT.ADTSigNotations.

(* Notations for Computational ADTs. *)

Bind Scope ADT_Scope with ADT.
Delimit Scope ADT_scope with ADT.

Import ListNotations.

(* Notations for computational ADT methods. *)

Class CoDHint := {codHint : option Type}.
Class RepHint := {rep : Type}.

Record cMethDef {Rep : Type} (Sig : methSig) :=
{ methBody :> methodType (methArity Sig) Rep (methDom Sig) (methCod Sig)}.

Arguments methBody [_ _] _.

Notation "'Def' 'Method0' id r .. xn : 'rep' '*' cod := bod" :=
  (Build_cMethDef _
     {| methID := id; methArity := 1; methDom := nil ; methCod := Some (cod : Type) |}
     (fun r => .. (fun xn =>
                     let _ := {| codHint := Some (cod : Type) |} in
                     (bod : methodType' rep [ ] codHint )) ..))
    (no associativity, id at level 0, r closed binder , xn closed binder,
     only parsing,
     at level 94,
     format "'Def'  'Method0'  id  r  ..  xn  :  'rep'  '*'  cod  :=  '/' '[  '   bod ']' ")
  : cMethDefParsing_scope.

Notation "'Def' 'Method1' id r .. xn : 'rep' '*' cod := bod" :=
  (Build_cMethDef _
     {| methID := id; methArity := 1; methDom := cons _ nil ; methCod := Some (cod : Type) |}
     (fun r => .. (fun xn =>
                     let _ := {| codHint := Some (cod : Type) |} in
                     (bod : methodType' rep [ ] codHint )) ..))
    (no associativity, id at level 0, r closed binder , xn closed binder,
     only parsing,
     at level 94,
     format "'Def'  'Method1'  id  r  ..  xn  :  'rep'  '*'  cod  :=  '/' '[  '   bod ']' ")
  : cMethDefParsing_scope.

Notation "'Def' 'Method2' id r .. xn : 'rep' '*' cod := bod" :=
  (Build_cMethDef _
     {| methID := id; methArity := 1; methDom := [ _; _ ] ; methCod := Some (cod : Type) |}
     (fun r => .. (fun xn =>
                     let _ := {| codHint := Some (cod : Type) |} in
                     (bod : methodType' rep [ ] codHint )) ..))
    (no associativity, id at level 0, r closed binder , xn closed binder,
     only parsing,
     at level 94,
     format "'Def'  'Method2'  id  r  ..  xn  :  'rep'  '*'  cod  :=  '/' '[  '   bod ']' ")
  : cMethDefParsing_scope.

Notation "'Def' 'Method3' id r .. xn : 'rep' '*' cod := bod" :=
  (Build_cMethDef _
     {| methID := id; methArity := 1; methDom := [ _; _; _ ] ; methCod := Some (cod : Type) |}
     (fun r => .. (fun xn =>
                     let _ := {| codHint := Some (cod : Type) |} in
                     (bod : methodType' rep [ ] codHint )) ..))
    (no associativity, id at level 0, r closed binder , xn closed binder,
     only parsing,
     at level 94,
     format "'Def'  'Method3'  id  r  ..  xn  :  'rep'  '*'  cod  :=  '/' '[  '   bod ']' ")
  : cMethDefParsing_scope.

Notation "'Def' 'Method4' id r .. xn : 'rep' '*' cod := bod" :=
  (Build_cMethDef _
     {| methID := id; methArity := 1; methDom := [_; _; _; _ ] ; methCod := Some (cod : Type) |}
     (fun r => .. (fun xn =>
                     let _ := {| codHint := Some (cod : Type) |} in
                     (bod : methodType' rep [ ] codHint )) ..))
    (no associativity, id at level 0, r closed binder , xn closed binder,
     only parsing,
     at level 94,
     format "'Def'  'Method4'  id  r  ..  xn  :  'rep'  '*'  cod  :=  '/' '[  '   bod ']' ")
  : cMethDefParsing_scope.

Notation "'Def' 'Method5' id r .. xn : 'rep' '*' cod := bod" :=
  (Build_cMethDef _
     {| methID := id; methArity := 1; methDom := [_; _; _; _; _ ] ; methCod := Some (cod : Type) |}
     (fun r => .. (fun xn =>
                     let _ := {| codHint := Some (cod : Type) |} in
                     (bod : methodType' rep [ ] codHint )) ..))
    (no associativity, id at level 0, r closed binder , xn closed binder,
     only parsing,
     at level 94,
     format "'Def'  'Method5'  id  r  ..  xn  :  'rep'  '*'  cod  :=  '/' '[  '   bod ']' ")
  : cMethDefParsing_scope.

(* Variant Notations for methods that don't return a value. *)

Notation "'Def' 'Method0' id r .. xn : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 1; methDom := [ ] ; methCod := None |} (fun r => .. (fun xn => let _ := {| codHint := None |} in (bod : methodType' rep [ ] codHint )) ..))
    (no associativity, id at level 0, r closed binder , xn closed binder,
     only parsing,
     at level 94,
     format "'Def'  'Method0'  id  r  ..  xn  :  'rep' :=  '/' '[  '   bod ']' ")
  : cMethDefParsing_scope.

Notation "'Def' 'Method1' id r .. xn : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 1; methDom := [_ ] ; methCod := None |} (fun r => .. (fun xn => let _ := {| codHint := None |} in (bod : methodType' rep [ ] codHint )) ..))
    (no associativity, id at level 0, r closed binder , xn closed binder,
     only parsing,
     at level 94,
     format "'Def'  'Method1'  id  r  ..  xn  :  'rep' :=  '/' '[  '   bod ']' ")
  : cMethDefParsing_scope.

Notation "'Def' 'Method2' id r .. xn : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 1; methDom := [_; _ ] ; methCod := None |} (fun r => .. (fun xn => let _ := {| codHint := None |} in (bod : methodType' rep [ ] codHint )) ..))
    (no associativity, id at level 0, r closed binder , xn closed binder,
     only parsing,
     at level 94,
     format "'Def'  'Method2'  id  r  ..  xn  :  'rep' :=  '/' '[  '   bod ']' ")
  : cMethDefParsing_scope.

Notation "'Def' 'Method3' id r .. xn : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 1; methDom := [_; _; _ ] ; methCod := None |} (fun r => .. (fun xn => let _ := {| codHint := None |} in (bod : methodType' rep [ ] codHint )) ..))
    (no associativity, id at level 0, r closed binder , xn closed binder,
     only parsing,
     at level 94,
     format "'Def'  'Method3'  id  r  ..  xn  :  'rep' :=  '/' '[  '   bod ']' ")
  : cMethDefParsing_scope.

Notation "'Def' 'Method4' id r .. xn : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 1; methDom := [_; _; _; _] ; methCod := None |} (fun r => .. (fun xn => let _ := {| codHint := None |} in (bod : methodType' rep [ ] codHint )) ..))
    (no associativity, id at level 0, r closed binder , xn closed binder,
     only parsing,
     at level 94,
     format "'Def'  'Method4'  id  r  ..  xn  :  'rep' :=  '/' '[  '   bod ']' ")
  : cMethDefParsing_scope.

Notation "'Def' 'Method5' id r .. xn : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 1; methDom := [_; _; _; _; _] ; methCod := None |} (fun r => .. (fun xn => let _ := {| codHint := None |} in (bod : methodType' rep [ ] codHint )) ..))
    (no associativity, id at level 0, r closed binder , xn closed binder,
     only parsing,
     at level 94,
     format "'Def'  'Method5'  id  r  ..  xn  :  'rep' :=  '/' '[  '   bod ']' ")
  : cMethDefParsing_scope.

(* Again, pretty printing involves fewer rules. *)
Notation "'Def' 'Method' id ( r : 'rep' ) : 'rep' '*' cod := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 1; methDom := [] ; methCod := Some cod |} (fun r => bod ))
    (no associativity, id at level 0, r at level 0,
     cod at level 0,
     at level 94,
     format "'Def'  'Method'  id  ( r  :  'rep' )  :  'rep'  '*'  cod  :=  '/' '[  '   bod ']' ")
  : cMethDef_scope.

Notation "'Def' 'Method' id ( r : 'rep' ) x1 .. xn : 'rep' '*' cod := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 1; methDom := _ :: _ ; methCod := Some cod |} (fun r =>
                                                                             (fun x1 => .. (fun xn => (bod )) ..)))
    (no associativity, id at level 0, r at level 0, x1 closed binder , xn closed binder,
     at level 94,
     format "'Def'  'Method'  id  ( r  :  'rep' )  x1  ..  xn  :  'rep'  '*'  cod  :=  '/' '[  '   bod ']' ")
  : cMethDef_scope.

Notation "'Def' 'Method' id ( r : 'rep' ) : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 1; methDom := [] ; methCod := None |} (fun r => bod ))
    (no associativity, id at level 0, r at level 0,
     at level 94,
     format "'Def'  'Method'  id  ( r  :  'rep' )  :  'rep'  :=  '/' '[  '   bod ']' ")
  : cMethDef_scope.

Notation "'Def' 'Method' id ( r : 'rep' ) x1 .. xn : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 1; methDom := _ :: _ ; methCod := None |} (fun r =>
                                                                         (fun x1 => .. (fun xn => (bod )) ..)))
    (no associativity, id at level 0, r at level 0, x1 closed binder , xn closed binder,
     at level 94,
     format "'Def'  'Method'  id  ( r  :  'rep' )  x1  ..  xn  :  'rep'  :=  '/' '[  '   bod ']' ")
  : cMethDef_scope.

Delimit Scope cMethDefParsing_scope with cMethDefParsing.
Delimit Scope cMethDef_scope with cMethDef.

(* Notations for parsing Constructors. Including the arity is the simplest way to
 make typechecking work. *)
Notation "'Def' 'Constructor0' id : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 0; methDom := []; methCod := None |} ((bod : constructorType rep [ ]) ))
    (no associativity, at level 94, id at level 0,
     format "'Def'  'Constructor0'  id  :  'rep'  :=  '/' '[  '   bod ']' " )
  : cMethDefParsing_scope.

Notation "'Def' 'Constructor1' id x1 .. xn : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 0; methDom := [_] ; methCod := None |} ((fun x1 => .. ((fun xn => (bod : constructorType rep [ ]) )) ..)))
    (no associativity, at level 94, id at level 0,
     x1 closed binder, xn closed binder,
     format "'Def'  'Constructor1'  id  x1  ..  xn  :  'rep'  :=  '/' '[  '   bod ']' " )
  : cMethDefParsing_scope.

Notation "'Def' 'Constructor2' id x1 .. xn : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 0; methDom := [_; _] ; methCod := None |} ((fun x1 => .. ((fun xn => (bod : constructorType rep [ ]) )) ..)))
    (no associativity, at level 94, id at level 0,
     x1 closed binder, xn closed binder,
     format "'Def'  'Constructor2'  id  x1  ..  xn  :  'rep'  :=  '/' '[  '   bod ']' " )
  : cMethDefParsing_scope.

Notation "'Def' 'Constructor3' id x1 .. xn : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 0; methDom := [_; _; _] ; methCod := None |} ((fun x1 => .. ((fun xn => (bod : constructorType rep [ ]) )) ..)))
    (no associativity, at level 94, id at level 0,
     x1 closed binder, xn closed binder,
     format "'Def'  'Constructor3'  id  x1  ..  xn  :  'rep'  :=  '/' '[  '   bod ']' " )
  : cMethDefParsing_scope.

Notation "'Def' 'Constructor4' id x1 .. xn : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 0; methDom := [_; _; _ ; _] ; methCod := None |} ((fun x1 => .. ((fun xn => (bod : constructorType rep [ ]) )) ..)))
    (no associativity, at level 94, id at level 0,
     x1 closed binder, xn closed binder,
     format "'Def'  'Constructor4'  id  x1  ..  xn  :  'rep'  :=  '/' '[  '   bod ']' " )
  : cMethDefParsing_scope.

Notation "'Def' 'Constructor5' id x1 .. xn : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 0; methDom := [_; _; _; _; _] ; methCod := None |} ((fun x1 => .. ((fun xn => (bod : constructorType rep [ ]) )) ..)))
    (no associativity, at level 94, id at level 0,
     x1 closed binder, xn closed binder,
     format "'Def'  'Constructor5'  id  x1  ..  xn  :  'rep'  :=  '/' '[  '   bod ']' " )
  : cMethDefParsing_scope.


(* Once the type infomation is in hand, two rules can handle pretty printing.  *)
Notation "'Def' 'Constructor' id x1 .. xn : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 1; methDom := _ ; methCod := None |} ((fun x1 => .. ((fun xn => bod )) ..)))
    (no associativity, at level 94, id at level 0,
     x1 closed binder, xn closed binder,
     format "'Def'  'Constructor'  id  x1  ..  xn  :  'rep'  :=  '/' '[  '   bod ']' " ) :
    cMethDef_scope.

Notation "'Def' 'Constructor' id : 'rep' := bod" :=
  (Build_cMethDef _ {| methID := id; methArity := 1; methDom := [] ; methCod := None |} bod)
    (no associativity, at level 94, id at level 0,
     format "'Def'  'Constructor'  id :  'rep'  :=  '/' '[  '   bod ']' " ) :
    cMethDef_scope.

Delimit Scope cMethDef_scope with cMethDef.

(* Lookup functions for constructor and method definitions. Because
   these definitions are parameterized on a signature, their
   bodies are contained in an indexed list [ilist] which is
   parameterized on a list of those signatures. *)

Definition getcMethDef
           (Rep : Type)
           {n}
           (methSigs : Vector.t methSig n)
           (methDefs : ilist methSigs)
           (idx : Fin.t n)
  : methodType
      (methArity (Vector.nth methSigs idx))
      Rep
      (methDom (Vector.nth methSigs idx))
      (methCod (Vector.nth methSigs idx)) :=
  methBody (ith methDefs idx).

(* Always simplify method lookup when the index is specified. *)
Arguments getcMethDef [_] {n} [_] _ idx%string / .

(* [BuildcADT] constructs an computational ADT from a single constructor
   definition and a list of method signatures,
   both indexed by their signatures. *)

Definition DecoratedADT (dSig : DecoratedADTSig) := ADT dSig.

Program Definition BuildADT
        {Rep : Type}
        {n'}
        {methSigs : Vector.t methSig n'}
        (methDefs : ilist (B:= @cMethDef Rep) methSigs)
: DecoratedADT (BuildADTSig methSigs)
      := existT _ Rep {|
                  pcMethods idx := getcMethDef methDefs idx
                |}.

(* Notation for ADTs built from [BuildADT]. *)

Notation "'Def' 'ADT' { 'rep' ':=' r , meth1 , .. , methn } " :=
  (let _ := {| rep := r |} in
    @BuildADT r
             _ _
             (icons meth1%cMethDefParsing .. (icons methn%cMethDefParsing (@inil _ (@cMethDef r))) ..))
    (no associativity, at level 96, r at level 0,
     format "'Def'  'ADT'   '/' '[hv  ' {  'rep'  ':=' r  , '//' meth1 , '//' .. , '//' methn  ']' }") :
    ADTParsing_scope.

Notation "'Def' 'ADT' { 'rep' ':=' r , meth1 , .. , methn } " :=
  (@BuildADT r
             _ _
             (icons meth1%cMethDef .. (icons methn%cMethDef (@inil _  (@cMethDef r))) ..))
    (no associativity, at level 96, r at level 0,
     format "'Def'  'ADT'  '/' '[hv  ' {  'rep'  ':='  r  ,  meth1 , '//' .. , '//' methn  ']' }") : ADT_scope.

(* Notations for method calls. *)

Definition callADTMethod
           {dSig : DecoratedADTSig}
           {rep : Type}
           (adt : pcADT dSig rep)
           (idxMap : BoundedLookup.BoundedIndex (MethodNames dSig) -> MethodIndex dSig)
           (idx : BoundedLookup.BoundedIndex (MethodNames dSig))
  := pcMethods adt (idxMap idx).

Notation "CompADT --> idx" := (callADTMethod CompADT (fun idx => BoundedLookup.ibound (BoundedLookup.indexb idx))
{| BoundedLookup.bindex := idx |}) (at level 10).

Notation CallMethod CompADT idx := (callADTMethod CompADT (fun idx => ibound (indexb idx))
{| bindex := idx |}).
