Require Import
        ADT.IterateBoundedIndex
        ADT.ADT
        ADT.ADTSigNotations
        ADT.ADTNotations.

Section MethodCompatibility.

  (** Old and new representations **)
  Variables oldRep newRep : Type.

  (** Abstraction Relation *)
  Variable AbsR : oldRep -> newRep -> Prop.

  Notation "ro ≃ rn" := (AbsR ro rn) (at level 70).

  (** Refinement of a constructor: the computation produced by
   a constructor [newConstructor] should be a refinement
   of the computation produced by the old constructor [oldObserver] on
   [d].  That is, the following diagram should commute:
<<
           old constructor
       Dom --------------> old rep
        ∥                      |
        ∥ id              AbsR |
        ∥                      |
       Dom --------------> new rep
          new constructor
>>
   *)

  Fixpoint compatibleMethod'
           {dom : list Type}
           {cod : option Type}
    : methodType' oldRep dom cod
      -> methodType' newRep dom cod
      -> Prop :=
    match dom return
          methodType' oldRep dom cod
          -> methodType' newRep dom cod
          -> Prop
    with
    | nil =>
      match cod return
            methodType' oldRep nil cod
            -> methodType' newRep nil cod
            -> Prop
      with
      | Some cod' =>
        fun oldMethod newMethod =>
          (fst oldMethod  ≃ fst newMethod) /\
          (snd oldMethod = snd newMethod)
      | _ =>
        fun oldMethod newMethod =>
          (oldMethod  ≃ newMethod)
      end
    | cons D dom' =>
      fun oldMethod newMethod =>
        forall d : D,
          @compatibleMethod' dom' cod (oldMethod d)
                        (newMethod d)
    end.

  Fixpoint compatibleMethod
           (arity : nat)
           {dom : list Type}
           {cod : option Type}
           (oldMethod : methodType arity oldRep dom cod)
           (newMethod : methodType arity newRep dom cod)
    : Prop
    :=
      match arity return methodType arity oldRep dom cod
                         -> methodType arity newRep dom cod
                         -> Prop with
      | 0 => @compatibleMethod' dom cod
      | S arity' =>
        fun oldMethod newMethod =>
          forall r_o r_n,
            r_o ≃ r_n ->
            compatibleMethod arity' (oldMethod r_o) (newMethod r_n)
      end oldMethod newMethod.

End MethodCompatibility.

Definition BisimiliarADT {Sig} (A B : ADT Sig) AbsR :=
  forall idx : MethodIndex Sig,
    @compatibleMethod
      (projT1 A) (projT1 B) AbsR
      (fst (fst (MethodDomCod Sig idx)))
      (snd (fst (MethodDomCod Sig idx)))
      (snd (MethodDomCod Sig idx))
      (pcMethods (projT2 A) idx)
      (pcMethods (projT2 B) idx).

Lemma proveBisimilarADT
      {n}
      {methSigs : Vector.t methSig n}
      (dSig := BuildADTSig methSigs)
      (A B : DecoratedADT dSig)
           AbsR
  : BisimiliarADT A B AbsR <->
    Iterate_Ensemble_BoundedIndex
      _ (fun idx : Fin.t (NumMethods dSig) =>
           @compatibleMethod
             (projT1 A) (projT1 B) AbsR
             (fst (fst (MethodDomCod dSig idx)))
             (snd (fst (MethodDomCod dSig idx)))
             (snd (MethodDomCod dSig idx))
             (pcMethods (projT2 A) idx)
             (pcMethods (projT2 B) idx)).
Proof.
  unfold BisimiliarADT; split; intros.
  eapply Iterate_Ensemble_BoundedIndex_equiv; eauto.
  pattern idx; eapply Iterate_Ensemble_BoundedIndex_equiv; eauto.
Qed.

Ltac start_bisimilarity_proof :=
  apply proveBisimilarADT; simpl; repeat apply IterateBoundedIndex.Build_prim_and;
  try exact I.
