Require Import Coq.Arith.Arith
        Coq.Lists.List
        Coq.Sets.Ensembles
        Coq.Strings.String
        ADT.ilist
        ADT.BoundedLookup.

Fixpoint app_assoc {A : Set} (As As' As'' : list A):
  As ++ (As' ++ As'') = (As ++ As') ++ As'' :=
  match As as As return
        As ++ (As' ++ As'') = (As ++ As') ++ As'' with
    | nil => eq_refl _
    | a :: As => (f_equal (fun l => a :: l) (app_assoc As As' As''))
  end.

Lemma app_comm_cons' {A : Set}:
  forall (a : A) As As',
    As ++ (a :: As') = (As ++ (a :: nil)) ++ As'.
Proof.
  intros; rewrite <- app_assoc; reflexivity.
Defined.

Section Iterate_Ensemble.

  Record prim_and (A B : Prop) : Prop :=
    { prim_fst : A;
      prim_snd : B }.

  Local Notation "A '/\' B" := (prim_and A B).

  Fixpoint Iterate_Ensemble_BoundedIndex'
           (Remaining : nat)
           (P : Ensemble (Fin.t Remaining))
           {struct Remaining}
    : Prop :=
    match Remaining return
          Ensemble (Fin.t Remaining)
                          -> Prop with
    | 0 => fun  P => True
    | S Remaining' =>
      fun P =>
        P (@Fin.F1 Remaining')
        /\ @Iterate_Ensemble_BoundedIndex' Remaining' (fun n' => P (@Fin.FS _ n'))
    end  P.

  Lemma Iterate_Ensemble_equiv'
    : forall (Remaining  : nat)
             (P : Ensemble (Fin.t Remaining)),
      (@Iterate_Ensemble_BoundedIndex'  Remaining P ->
       forall idx, P idx).
  Proof.
    induction Remaining; simpl; intros.
    - inversion idx.
    - revert IHRemaining  P H.
      pattern Remaining, idx; apply Fin.caseS; intros.
      + intuition.
      + intuition; eapply (IHRemaining _ prim_snd0).
  Qed.

  Lemma Iterate_Ensemble_equiv''
    : forall (Remaining  : nat)
             (P : Ensemble (Fin.t Remaining)),
      (forall idx, P idx)
      -> @Iterate_Ensemble_BoundedIndex'  Remaining P.
  Proof.
    induction Remaining; simpl; intros; intuition.
  Qed.

  Definition Iterate_Ensemble_BoundedIndex
             (m : nat)
             (P : Ensemble (Fin.t m)) : Prop :=
    @Iterate_Ensemble_BoundedIndex' m P.

  Corollary Iterate_Ensemble_BoundedIndex_equiv
  : forall m
           (P : Ensemble (Fin.t m)),
      Iterate_Ensemble_BoundedIndex _ P <->
      forall idx, P idx.
  Proof.
    split; intros.
    eapply Iterate_Ensemble_equiv';
      eauto using string_dec; destruct n; simpl; discriminate.
    eapply Iterate_Ensemble_equiv'';
      eauto using string_dec.
  Qed.

  (* Iterating with a filter *)
  Fixpoint Iterate_Ensemble_BoundedIndex_filter'
           ( Remaining : nat)
           (filter : Fin.t Remaining -> bool)
           (P : Ensemble (Fin.t Remaining))
           {struct Remaining}
    : Prop :=
    match Remaining return
            (@Fin.t Remaining -> bool)
            -> Ensemble (Fin.t Remaining)
                          -> Prop with
    | 0 => fun  filter P => True
    | S Remaining' =>
      fun  filter P =>
        if filter (@Fin.F1 Remaining') then
          P (@Fin.F1 Remaining')
          /\ @Iterate_Ensemble_BoundedIndex_filter' Remaining'
               (fun n' => filter (@Fin.FS _ n'))
               (fun n' => P (@Fin.FS _ n'))
        else
          @Iterate_Ensemble_BoundedIndex_filter' Remaining'
            (fun n' => filter (@Fin.FS _ n'))
            (fun n' => P (@Fin.FS _ n'))
    end  filter P.


End Iterate_Ensemble.

(* Always expand these iterations. *)
Arguments Iterate_Ensemble_BoundedIndex _ _ / .

Section Iterate_Dep_Type.

  Local Notation Dep_Type A := (A -> Type).

  Fixpoint Iterate_Dep_Type_BoundedIndex'
           (Remaining : nat)
           (P : Dep_Type (Fin.t Remaining))
           {struct Remaining}
    : Type :=
    match Remaining return
           Dep_Type (Fin.t Remaining)
                          -> Type with
    | 0 => fun  P => unit
    | S Remaining' =>
      fun  P =>
        prim_prod (P (@Fin.F1 Remaining'))
                  (@Iterate_Dep_Type_BoundedIndex'
                      Remaining'
                     (fun n' => P (@Fin.FS _ n')))
    end  P.

  Fixpoint Iterate_Dep_Type_equiv'
           (Remaining : nat)
           (P : Dep_Type (Fin.t Remaining))
           (H : @Iterate_Dep_Type_BoundedIndex' Remaining P)
           idx
           {struct idx}
    : P idx :=
    match idx in Fin.t n return
          forall (P : Dep_Type (Fin.t n)),
            @Iterate_Dep_Type_BoundedIndex' n P
            -> P idx with
    | Fin.F1 => fun P H => ilist.prim_fst H
    | Fin.FS n' =>
      fun P H =>
        Iterate_Dep_Type_equiv' _ (fun i => P (Fin.FS i))
                                (ilist.prim_snd H) n'
    end P H.

  Fixpoint Iterate_Dep_Type_equiv''
           (Remaining  : nat)
    : forall (P : Dep_Type (Fin.t Remaining)),
      (forall idx, P idx)
      -> @Iterate_Dep_Type_BoundedIndex' Remaining P :=
    match Remaining return
          forall (P : Dep_Type (Fin.t Remaining)),
            (forall idx, P idx)
            -> @Iterate_Dep_Type_BoundedIndex' Remaining P with
    | 0 => fun P H => tt
    | S n' =>
      fun P H =>
        Build_prim_prod (H Fin.F1)
                        (Iterate_Dep_Type_equiv'' _ (fun i => P (Fin.FS i))
                                                  (fun i => H (Fin.FS i)))

    end.

  Definition Iterate_Dep_Type_BoundedIndex
             (m : nat)
             (P : Dep_Type (Fin.t m)) : Type :=
    @Iterate_Dep_Type_BoundedIndex' m P.

  Corollary Lookup_Iterate_Dep_Type
  : forall m
           (P : Dep_Type (Fin.t m)),
      Iterate_Dep_Type_BoundedIndex _ P ->
      forall idx, P idx.
  Proof.
    intros; eapply Iterate_Dep_Type_equiv' ; eassumption.
  Defined.

  Corollary Iterate_Dep_Type_BoundedIndex_equiv2
    : forall m
             (P : Dep_Type (Fin.t m)),
      (forall idx, P idx)
      -> Iterate_Dep_Type_BoundedIndex _ P.
  Proof.
    intros.
    eapply Iterate_Dep_Type_equiv'';
      eauto using string_dec.
  Defined.

  (* Iterating with a filter *)

  Fixpoint Update_Iterate_Dep_Type
           (n : nat)
           (k : Fin.t n)
           (P : Dep_Type (Fin.t n))
           (X0 : Iterate_Dep_Type_BoundedIndex _ P)
           (a' : P k )
           {struct k}
  : Iterate_Dep_Type_BoundedIndex _ P :=
    match k in Fin.t n return
          forall (P : Fin.t n -> Type),
            Iterate_Dep_Type_BoundedIndex _ P
            -> P k -> Iterate_Dep_Type_BoundedIndex _ P with
    | Fin.F1 => fun P X0 a' => Build_prim_prod a' (ilist.prim_snd X0)
    | Fin.FS k' => fun P X0 a' => Build_prim_prod (ilist.prim_fst X0)
                                                    (Update_Iterate_Dep_Type _ k'
                                                                             (fun i => P (Fin.FS i)) (ilist.prim_snd X0) a')
    end P X0 a'.

  Definition Lookup_Update_Iterate_Dep_Type_eq {n}
    : forall (k : Fin.t n)
             (P : Fin.t n -> Type)
             (X0 : Iterate_Dep_Type_BoundedIndex _ P)
             (a' : P k ),
      Lookup_Iterate_Dep_Type _ P (Update_Iterate_Dep_Type _ k P X0 a') k = a'.
  Proof.
    induction k; simpl; intros.
    - eauto.
    - rewrite IHk; eauto.
  Qed.

  Definition Lookup_Update_Iterate_Dep_Type_neq {n}
    : forall (k k': Fin.t n)
             (P : Fin.t n -> Type)
             (X0 : Iterate_Dep_Type_BoundedIndex _ P)
             (a' : P k),
      k' <> k
      -> Lookup_Iterate_Dep_Type _ P (Update_Iterate_Dep_Type _ k P X0 a') k' =
         Lookup_Iterate_Dep_Type _ P X0 k'.
  Proof.
    induction k; simpl.
    - intro k'; pattern n, k'.
      match goal with
        |- ?P n k' => simpl; eapply (Fin.caseS P); simpl
      end; congruence.
    - intro k'; revert k IHk; pattern n, k'.
      match goal with
        |- ?P n k' => simpl; eapply (Fin.caseS P); simpl
      end; try congruence.
      intros.
      rewrite IHk; congruence.
  Qed.

End Iterate_Dep_Type.

(* Always expand these iterations as well. *)
Arguments Iterate_Dep_Type_BoundedIndex _ _ / .
