(****** Homework 8 for CS-565 Programming Languages - Purdue. ******)
(** !!!Due date: 12/7/2016 by 9:00PM **)
(* All exercises are to be completed.  *)

Require Import Coq.Lists.List.
Import ListNotations.
Require Import SfLib.
Require Import Maps.
Require Import Smallstep.
Require Import Norm. (* Language Definitions for Normalization Proof. *)
Hint Constructors multi.

(* ============== Normalization ==============*)
(** Your goal for the first part of this assignement is to prove that
every well-typed term for the simply typed lambda calculus with booleans
and pairs (defined in Norm.v) halts, i.e. reduces to a value.  *)

Definition halts  (t:tm) : Prop :=  exists t', t ==>* t' /\  value t'.

Lemma value_halts : forall v, value v -> halts v.
Proof.
  intros v H. unfold halts.
  exists v. split.
  apply multi_refl.
  assumption.
Qed.

(** The key issue in the normalization proof (as in many proofs by
    induction) is finding a strong enough induction hypothesis.  To
    this end, we begin by defining, for each type [T], a set [R_T] of
    closed terms of type [T] i.e. have type T in the empty context.
    We will specify these sets using a relation [R] and write [R T t]
    when [t] is in [R_T].

    Here is the definition of [R] for our language:

    - [R bool t] iff [t] is a closed term of type [bool] and [t] halts
      in a value

    - [R (T1 -> T2) t] iff [t] is a closed term of type [T1 -> T2] and
      [t] halts in a value _and_ for any term [s] such that [R T1 s],
      we have [R T2 (t s)].

    - [R (T1 * T2) t] iff [t] is a closed term of type [T1 * T2],
      [t] halts in a value _and_ the components of t belong to the relation:
      [R T1 (fst t)] and [R T2 (snd t)].
*)

(** This definition gives us the strengthened induction hypothesis that we
    need.  Our primary goal is to show that all _programs_ ---i.e., all
    closed terms of base type---halt.  But closed terms of base type can
    contain subterms of functional type, so we need to know something
    about these as well.  Moreover, it is not enough to know that these
    subterms halt, because the application of a normalized function to a
    normalized argument involves a substitution, which may enable more
    reduction steps.  So we need a stronger condition for terms of
    functional type: not only should they halt themselves, but, when
    applied to halting arguments, they should yield halting results.

    The form of [R] is characteristic of the _logical relations_ proof
    technique that we previously saw in class.  If we want to
    prove some property [P] of all closed terms of type [A], we proceed by
    proving, by induction on types, that all terms of type [A] _possess_
    property [P], all terms of type [A->A] _preserve_ property [P], all
    terms of type [(A->A)->(A->A)] _preserve the property of preserving_
    property [P], and so on.  We do this by defining a family of
    properties, indexed by types.  For the base type [A], the property is
    just [P].  For functional types, it says that the function should map
    values satisfying the property at the input type to values satisfying
    the property at the output type.

    Fortunately, we define [R] using a  [Fixpoint]: *)


(** **** Exercise: 2 stars  Update the definition of R with the correct *)
(* definition for product types. *)

Fixpoint R (T:ty) (t:tm) {struct T} : Prop :=
  has_type empty t T /\ halts t /\
  (match T with
   | TBool  => True
   | TArrow T1 T2 => (forall s, R T1 s -> R T2 (tapp t s))
   | TProd T1 T2 => False (* Fill in Here. *)
   end).

(** As immediate consequences of this definition, we have that every
    element of every set [R_T] halts in a value and is closed with type
    [t] :*)

Lemma R_halts : forall {T} {t}, R T t -> halts t.
Proof.
  intros. destruct T; unfold R in H; inversion H; inversion H1;  assumption.
Qed.


Lemma R_typable_empty : forall {T} {t}, R T t -> has_type empty t T.
Proof.
  intros. destruct T; unfold R in H; inversion H; inversion H1; assumption.
Qed.

(** Now we proceed to show the main result, which is that every
    well-typed term of type [T] is an element of [R_T].  Together with
    [R_halts], that will show that every well-typed term halts in a
    value.  *)


(* ================================================================= *)
(** **  Membership in [R_T] Is Invariant Under Reduction *)

(** We start with a preliminary lemma that shows a kind of strong
    preservation property, namely that membership in [R_T] is _invariant_
    under reduction. We will need this property in both directions,
    i.e., both to show that a term in [R_T] stays in [R_T] when it takes a
    forward step, and to show that any term that ends up in [R_T] after a
    step must have been in [R_T] to begin with.

    First of all, an easy preliminary lemma. Note that in the forward
    direction the proof depends on the fact that our language is
    determinstic. This lemma might still be true for nondeterministic
    languages, but the proof would be harder! *)

Lemma step_preserves_halting : forall t t', (t ==> t') -> (halts t <-> halts t').
Proof.
 intros t t' ST.  unfold halts.
 split.
 - (* -> *)
  intros [t'' [STM V]].
  inversion STM; subst.
   exfalso.  apply value__normal in V. unfold normal_form in V. apply V. exists t'. auto.
   rewrite (step_deterministic _ _ _ ST H). exists t''. split; assumption.
 - (* <- *)
  intros [t'0 [STM V]].
  exists t'0. split; eauto.
Qed.

(** Now the main lemma, which comes in two parts, one for each
    direction.  Each proceeds by induction on the structure of the type
    [T]. In fact, this is where we make fundamental use of the
    structure of types.

    One requirement for staying in [R_T] is to stay in type [T]. In the
    forward direction, we get this from ordinary type Preservation. *)

(** **** Exercise: 3 stars, Fix step_preserves_R for your definition of R.  *)
Lemma step_preserves_R : forall T t t', (t ==> t') -> R T t -> R T t'.
Proof.
 induction T;  intros t t' E Rt; unfold R; fold R; unfold R in Rt; fold R in Rt;
               destruct Rt as [typable_empty_t [halts_t RRt]].
  (* TBool *)
  split. eapply preservation; eauto.
  split. apply (step_preserves_halting _ _ E); eauto.
  auto.
  (* TArrow *)
  split. eapply preservation; eauto.
  split. apply (step_preserves_halting _ _ E); eauto.
  intros.
  eapply IHT2.
  apply  ST_App1. apply E.
  apply RRt; auto.
  (* TProd *)
  (* Fill In Here *)
Admitted.

(** The generalization to multiple steps is trivial: *)
Lemma multistep_preserves_R : forall T t t',
  (t ==>* t') -> R T t -> R T t'.
Proof.
  intros T t t' STM; induction STM; intros.
  assumption.
  apply IHSTM. eapply step_preserves_R. apply H. assumption.
Qed.

Lemma Uniqueness_of_types :
  forall t T,
    has_type empty t T
    -> forall T',
      has_type empty t T'
      -> T = T'.
Proof.
  intros t T WF_t; induction WF_t; intros T' WF_t';
    inversion WF_t'; subst; try f_equal; eauto.
  - congruence.
  - apply IHWF_t1 in H2; congruence.
  - apply IHWF_t in H1; congruence.
  - apply IHWF_t in H1; congruence.
Qed.


(** **** Exercise: 2 stars, Prove the reverse direction of step_preserves_R.  *)
(* You may find [Uniqueness_of_types] useful.*)

Lemma step_preserves_R' : forall T t t',
  has_type empty t T -> (t ==> t') -> R T t' -> R T t.
Proof.
Admitted.

Lemma multistep_preserves_R' : forall T t t',
  has_type empty t T -> (t ==>* t') -> R T t' -> R T t.
Proof.
  intros T t t' HT STM.
  induction STM; intros.
    assumption.
    eapply step_preserves_R'.  assumption. apply H. apply IHSTM.
    eapply preservation;  eauto. auto.
Qed.

(* ================================================================= *)
(** ** Closed Instances of Terms of Type [t] Belong to [R_T] *)

(** Now we proceed to show that every term of type [T] belongs to
    [R_T].  Here, the induction will be on typing derivations (it would be
    surprising to see a proof about well-typed terms that did not
    somewhere involve induction on typing derivations!).  The only
    technical difficulty here is in dealing with the abstraction case.
    Since we are arguing by induction, the demonstration that a term
    [tabs x T1 t2] belongs to [R_(T1->T2)] should involve applying the
    induction hypothesis to show that [t2] belongs to [R_(T2)].  But
    [R_(T2)] is defined to be a set of _closed_ terms, while [t2] may
    contain [x] free, so this does not make sense.

    This problem is resolved by using a standard trick to suitably
    generalize the induction hypothesis: instead of proving a statement
    involving a closed term, we generalize it to cover all closed
    _instances_ of an open term [t].  Informally, the statement of the
    lemma will look like this:

    If [x1:T1,..xn:Tn |- t : T] and [v1,...,vn] are values such that
    [R T1 v1], [R T2 v2], ..., [R Tn vn], then
    [R T ([x1:=v1][x2:=v2]...[xn:=vn]t)].

    The proof will proceed by induction on the typing derivation
    [x1:T1,..xn:Tn |- t : T]; the most interesting case will be the one
    for abstraction. *)

(* ----------------------------------------------------------------- *)
(** *** Multisubstitutions, Multi-Extensions, and Instantiations *)

(** However, before we can proceed to formalize the statement and
    proof of the lemma, we'll need to build some (rather tedious)
    machinery to deal with the fact that we are performing _multiple_
    substitutions on term [t] and _multiple_ extensions of the typing
    context.  In particular, we must be precise about the order in which
    the substitutions occur and how they act on each other.  Often these
    details are simply elided in informal paper proofs, but of course Coq
    won't let us do that. Since here we are substituting closed terms, we
    don't need to worry about how one substitution might affect the term
    put in place by another.  But we still do need to worry about the
    _order_ of substitutions, because it is quite possible for the same
    identifier to appear multiple times among the [x1,...xn] with
    different associated [vi] and [Ti].

    To make everything precise, we will assume that environments are
    extended from left to right, and multiple substitutions are performed
    from right to left.  To see that this is consistent, suppose we have
    an environment written as [...,y:bool,...,y:nat,...]  and a
    corresponding term substitution written as [...[y:=(tbool
    true)]...[y:=(tnat 3)]...t].  Since environments are extended from
    left to right, the binding [y:nat] hides the binding [y:bool]; since
    substitutions are performed right to left, we do the substitution
    [y:=(tnat 3)] first, so that the substitution [y:=(tbool true)] has
    no effect. Substitution thus correctly preserves the type of the term.

    With these points in mind, the following definitions should make sense.

    A _multisubstitution_ is the result of applying a list of
    substitutions, which we call an _environment_. *)

Definition env := list (id * tm).

Fixpoint msubst (ss:env) (t:tm) {struct ss} : tm :=
match ss with
| nil => t
| ((x,s)::ss') => msubst ss' ([x:=s]t)
end.

(** We need similar machinery to talk about repeated extension of a
    typing context using a list of (identifier, type) pairs, which we
    call a _type assignment_. *)

Definition tass := list (id * ty).

Fixpoint mupdate (Gamma : context) (xts : tass) :=
  match xts with
  | nil => Gamma
  | ((x,v)::xts') => update (mupdate Gamma xts') x v
  end.

(** We will need some simple operations that work uniformly on
    environments and type assigments *)

Fixpoint lookup {X:Set} (k : id) (l : list (id * X)) {struct l}
              : option X :=
  match l with
    | nil => None
    | (j,x) :: l' =>
      if beq_id j k then Some x else lookup k l'
  end.

Fixpoint drop {X:Set} (n:id) (nxs:list (id * X)) {struct nxs}
            : list (id * X) :=
  match nxs with
    | nil => nil
    | ((n',x)::nxs') =>
        if beq_id n' n then drop n nxs'
        else (n',x)::(drop n nxs')
  end.

(** An _instantiation_ combines a type assignment and a value
    environment with the same domains, where corresponding elements are
    in R. *)

Inductive instantiation :  tass -> env -> Prop :=
| V_nil :
    instantiation nil nil
| V_cons : forall x T v c e,
    value v -> R T v ->
    instantiation c e ->
    instantiation ((x,T)::c) ((x,v)::e).

(** We now proceed to prove various properties of these definitions. *)

(* ----------------------------------------------------------------- *)
(** *** More Substitution Facts *)

(** First we need some additional lemmas on (ordinary) substitution. *)

Lemma vacuous_substitution : forall  t x,
     ~ appears_free_in x t  ->
     forall t', [x:=t']t = t.
Proof with eauto.
  induction t; simpl; intros...
  - destruct (beq_id x i) eqn: ?; eauto.
    apply beq_id_true_iff in Heqb; subst.
    exfalso...
  - rewrite IHt1, IHt2; eauto.
  - destruct (beq_id x i) eqn: ?; eauto.
    apply beq_id_false_iff in Heqb.
    rewrite IHt; eauto.
  - rewrite IHt1, IHt2; eauto.
  - rewrite IHt; eauto.
  - rewrite IHt; eauto.
  - rewrite IHt1, IHt2, IHt3; eauto.
Qed.

Lemma subst_closed: forall t,
     closed t  ->
     forall x t', [x:=t']t = t.
Proof.
  intros. apply vacuous_substitution. apply H.  Qed.

Lemma subst_not_afi : forall t x v,
    closed v ->  ~ appears_free_in x ([x:=v]t).
Proof with eauto.  (* rather slow this way *)
  unfold closed, not.
  induction t; intros x v P A; simpl in A.
    - (* tvar *)
     destruct (beq_idP x i)...
     inversion A; subst. auto.
    - (* tapp *)
     inversion A; subst...
    - (* tabs *)
     destruct (beq_idP x i)...
     + inversion A; subst...
     + inversion A; subst...
    - (* tpair *)
     inversion A; subst...
    - (* tfst *)
     inversion A; subst...
    - (* tsnd *)
     inversion A; subst...
    - (* ttrue *)
     inversion A.
    - (* tfalse *)
     inversion A.
    - (* tif *)
     inversion A; subst...
Qed.

Lemma duplicate_subst : forall t' x t v,
  closed v -> [x:=t]([x:=v]t') = [x:=v]t'.
Proof.
  intros. eapply vacuous_substitution. apply subst_not_afi.  auto.
Qed.

Lemma swap_subst : forall t x x1 v v1,
    x <> x1 ->
    closed v -> closed v1 ->
    [x1:=v1]([x:=v]t) = [x:=v]([x1:=v1]t).
Proof with eauto.
 induction t; intros; simpl.
  - (* tvar *)
   destruct (beq_idP x i); destruct (beq_idP x1 i).
   + subst. exfalso...
   + subst. simpl. rewrite <- beq_id_refl. apply subst_closed...
   + subst. simpl. rewrite <- beq_id_refl. rewrite subst_closed...
   + simpl. rewrite false_beq_id... rewrite false_beq_id...
  (* tapp *)
  - rewrite IHt1; eauto.
    rewrite IHt2; eauto.
  (* tabs *)
  - destruct (beq_idP x i); destruct (beq_idP x1 i)...
    rewrite IHt; eauto.
  (* tpair *)
  - rewrite IHt1; eauto.
    rewrite IHt2; eauto.
  (* tfst *)
  - rewrite IHt; eauto.
  (* tsnd *)
  -  rewrite IHt; eauto.
  (* ttrue *)
  - eauto.
  (* tfalse *)
  - eauto.
  (* tif *)
  - rewrite IHt1; eauto.
    rewrite IHt2; eauto.
    rewrite IHt3; eauto.
Qed.

(* ----------------------------------------------------------------- *)
(** *** Properties of Multi-Substitutions *)

Lemma msubst_closed: forall t, closed t -> forall ss, msubst ss t = t.
Proof.
  induction ss.
    reflexivity.
    destruct a. simpl. rewrite subst_closed; assumption.
Qed.

(** Closed environments are those that contain only closed terms. *)

Fixpoint closed_env (env:env) {struct env} :=
  match env with
  | nil => True
  | (x,t)::env' => closed t /\ closed_env env'
  end.

(** Next come a series of lemmas charcterizing how [msubst] of closed terms
    distributes over [subst] and over each term form *)

Lemma subst_msubst: forall env x v t, closed v -> closed_env env ->
    msubst env ([x:=v]t) = [x:=v](msubst (drop x env) t).
Proof.
  induction env0; intros; auto.
  destruct a. simpl.
  inversion H0. fold closed_env in H2.
  destruct (beq_idP i x).
  - subst. rewrite duplicate_subst; auto.
  - simpl. rewrite swap_subst; eauto.
Qed.

(** Exercise: 1 stars Fill in proofs for the other term constructors. *)
Lemma msubst_var:  forall ss x, closed_env ss ->
   msubst ss (tvar x) =
   match lookup x ss with
   | Some t => t
   | None => tvar x
  end.
Proof.
Admitted.

Lemma msubst_abs: forall ss x T t,
  msubst ss (tabs x T t) = tabs x T (msubst (drop x ss) t).
Proof.
Admitted.

Lemma msubst_app : forall ss t1 t2, msubst ss (tapp t1 t2) = tapp (msubst ss t1) (msubst ss t2).
Proof.
Admitted.

Lemma msubst_tpair : forall ss t1 t2, msubst ss (tpair t1 t2) = tpair (msubst ss t1) (msubst ss t2).
Proof.
Admitted.

Lemma msubst_tfst : forall ss t, msubst ss (tfst t) = tfst (msubst ss t).
Proof.
Admitted.

Lemma msubst_tsnd : forall ss t, msubst ss (tsnd t) = tsnd (msubst ss t).
Proof.
Admitted.

Lemma msubst_ttrue : forall ss, msubst ss ttrue = ttrue.
Proof.
Admitted.

Lemma msubst_tfalse : forall ss, msubst ss tfalse = tfalse.
Proof.
Admitted.

Lemma msubst_tif : forall ss t1 t2 t3,
    msubst ss (tif t1 t2 t3) = tif (msubst ss t1) (msubst ss t2) (msubst ss t3).
Proof.
Admitted.

(* ----------------------------------------------------------------- *)
(** *** Properties of Multi-Extensions *)

(** We need to connect the behavior of type assignments with that of
    their corresponding contexts. *)

Lemma mupdate_lookup : forall (c : tass) (x:id),
    lookup x c = (mupdate empty c) x.
Proof.
  induction c; intros.
    auto.
    destruct a. unfold lookup, mupdate, update, t_update. destruct (beq_id i x); auto.
Qed.

Lemma mupdate_drop : forall (c: tass) Gamma x x',
      mupdate Gamma (drop x c) x'
    = if beq_id x x' then Gamma x' else mupdate Gamma c x'.
Proof.
  induction c; intros.
  - destruct (beq_idP x x'); auto.
  - destruct a. simpl.
    destruct (beq_idP i x).
    + subst. rewrite IHc.
      unfold update, t_update. destruct (beq_idP x x'); auto.
    + simpl. unfold update, t_update. destruct (beq_idP i x'); auto.
      subst. rewrite false_beq_id; congruence.
Qed.

(* ----------------------------------------------------------------- *)
(** *** Properties of Instantiations *)

(** These are strightforward. *)

Lemma instantiation_domains_match: forall {c} {e},
    instantiation c e ->
    forall {x} {T},
      lookup x c = Some T -> exists t, lookup x e = Some t.
Proof.
  intros c e V. induction V; intros x0 T0 C.
    solve by inversion .
    simpl in *.
    destruct (beq_id x x0); eauto.
Qed.

Lemma instantiation_env_closed : forall c e,
  instantiation c e -> closed_env e.
Proof.
  intros c e V; induction V; intros.
    econstructor.
    unfold closed_env. fold closed_env.
    split.  eapply typable_empty__closed. eapply R_typable_empty. eauto.
        auto.
Qed.

Lemma instantiation_R : forall c e,
    instantiation c e ->
    forall x t T,
      lookup x c = Some T ->
      lookup x e = Some t -> R T t.
Proof.
  intros c e V. induction V; intros x' t' T' G E.
    solve by inversion.
    unfold lookup in *.  destruct (beq_id x x').
      inversion G; inversion E; subst.  auto.
      eauto.
Qed.

Lemma instantiation_drop : forall c env,
    instantiation c env ->
    forall x, instantiation (drop x c) (drop x env).
Proof.
  intros c e V. induction V.
    intros.  simpl.  constructor.
    intros. unfold drop. destruct (beq_id x x0); auto. constructor; eauto.
Qed.

(* ----------------------------------------------------------------- *)
(** *** We need to show if subterms of various terms reduce via multistep *)
(*  reduction, the whole term can as well. *)

Lemma multistep_App2 : forall v t t',
  value v -> (t ==>* t') -> (tapp v t) ==>* (tapp v t').
Proof.
  intros v t t' V STM. induction STM.
   apply multi_refl.
   eapply multi_step.
     apply ST_App2; eauto.  auto.
Qed.

(** Exercise: 1 stars Fill in proofs for the other term constructors. *)
Lemma multistep_tpair_2 : forall v t t',
  value v -> (t ==>* t') -> (tpair v t) ==>* (tpair v t').
Proof.
Admitted.

Lemma multistep_tpair_1 : forall t t' t2,
    (t ==>* t') -> (tpair t t2) ==>* (tpair t' t2).
Proof.
Admitted.

Lemma multistep_tfst : forall t t',
    (t ==>* t') -> (tfst t) ==>* (tfst t').
Proof.
Admitted.

Lemma multistep_tsnd : forall t t',
    (t ==>* t') -> (tsnd t) ==>* (tsnd t').
Proof.
Admitted.

Lemma multistep_tif : forall t t' t2 t3,
    (t ==>* t') -> (tif t t2 t3) ==>* (tif t' t2 t3).
Proof.
Admitted.

(* ----------------------------------------------------------------- *)
(** *** The R Lemma. *)

(** We can finally put everything together.

    The key lemma about preservation of typing under substitution can
    be lifted to multi-substitutions: *)

Lemma msubst_preserves_typing : forall c e,
     instantiation c e ->
     forall Gamma t S, has_type (mupdate Gamma c) t S ->
     has_type Gamma (msubst e t) S.
Proof.
  induction 1; intros.
    simpl in H. simpl. auto.
    simpl in H2.  simpl.
    apply IHinstantiation.
    eapply substitution_preserves_typing; eauto.
    apply (R_typable_empty H0).
Qed.

(** And at long last, the main lemma. *)

(** Exercise: 4 stars: Complete the missing cases to finish [msubst_R]. *)
Lemma msubst_R : forall c env t T,
    has_type (mupdate empty c) t T ->
    instantiation c env ->
    R T (msubst env t).
Proof.
  intros c env0 t T HT V.
  generalize dependent env0.
  (* We need to generalize the hypothesis a bit before setting up the induction. *)
  remember (mupdate empty c) as Gamma.
  assert (forall x, Gamma x = lookup x c).
    intros. rewrite HeqGamma. rewrite mupdate_lookup. auto.
  clear HeqGamma.
  generalize dependent c.
  induction HT; intros.
  - (* T_Var *)
   rewrite H0 in H. destruct (instantiation_domains_match V H) as [t P].
   eapply instantiation_R; eauto.
   rewrite msubst_var.  rewrite P. auto. eapply instantiation_env_closed; eauto.

  - (* T_Abs *)
    rewrite msubst_abs.
    (* We'll need variants of the following fact several times, so its simplest to
       establish it just once. *)
    assert (WT: has_type empty (tabs x T11 (msubst (drop x env0) t12)) (TArrow T11 T12)).
    { eapply T_Abs. eapply msubst_preserves_typing.
      { eapply instantiation_drop; eauto. }
      eapply context_invariance.
      { apply HT. }
      intros.
      unfold update, t_update. rewrite mupdate_drop. destruct (beq_idP x x0).
      + auto.
      + rewrite H.
        clear - c n. induction c.
        simpl.  rewrite false_beq_id; auto.
        simpl. destruct a.  unfold update, t_update.
        destruct (beq_id i x0); auto. }
    unfold R. fold R. split.
       auto.
     split. apply value_halts. apply v_abs.
     intros.
     destruct (R_halts H0) as [v [P Q]].
     pose proof (multistep_preserves_R _ _ _ P H0).
     apply multistep_preserves_R' with (msubst ((x,v)::env0) t12).
       eapply T_App. eauto.
       apply R_typable_empty; auto.
       eapply multi_trans.  eapply multistep_App2; eauto.
       eapply multi_R.
       simpl.  rewrite subst_msubst.
       eapply ST_AppAbs; eauto.
       eapply typable_empty__closed.
       apply (R_typable_empty H1).
       eapply instantiation_env_closed; eauto.
       eapply (IHHT ((x,T11)::c)).
          intros. unfold update, t_update, lookup. destruct (beq_id x x0); auto.
       constructor; auto.

  - (* T_App *)
    rewrite msubst_app.
    destruct (IHHT1 c H env0 V) as [_ [_ P1]].
    pose proof (IHHT2 c H env0 V) as P2.  fold R in P1.  auto.
  (* T_Pair *)
  - rewrite msubst_tpair.
    (* By the two induction hypotheses, we know that [R T1 (msubst env0 t1)] *)
    (* and [R T2 (msubst env0 t2)], which we can use to reduce *)
    (* [(tpair (msubst env0 t1) (msubst env0 t2))] to a value. If we can show that this
       value belongs to the relation [R (T1 * T2)], we can use [multistep_preserves_R'] to *)
    (* finish this case. *)
    admit.  (* Fill in Here. *)
  - rewrite msubst_tfst.
    (* This case should follow immediately from the induction hypothesis and
       your definition of R. *)
    admit.
  - (* This case should _also_ follow immediately from the induction hypothesis and
       your definition of R. *)
    admit.
  - rewrite msubst_ttrue; simpl; repeat split; eauto.
    unfold halts; eauto.
  - rewrite msubst_tfalse; simpl; repeat split; eauto.
    unfold halts; eauto.
  - rewrite msubst_tif.
    (* By IHHT1, [msubst env0 t0] will reduce to a value of type TBool. *)
    (* This tells us that [tif (msubst env0 t0) (msubst env0 t1) (msubst env0 t2)] *)
    (* will reduce to either [msubst env0 t1] or [msubst env0 t2], depending on *)
    (* whether that value is ttrue or tfalse. Use that fact, plus [multistep_preserves_R'] *)
    (* to complete this last case. *)
    admit.
Admitted.

(** *** Normalization Theorem *)
(* Thanks to all your hard work, the proof is done! *)
Theorem normalization : forall t T, has_type empty t T -> halts t.
Proof.
  intros.
  replace t with (msubst nil t) by reflexivity.
  apply (@R_halts T).
  apply (msubst_R nil); eauto.
  eapply V_nil.
Qed.

(* ============== Curry Howard Isomorphism ============== *)

(** **** Exercise: 1 star (eight_is_even)  *)
(** Give a proof object showing that [ev 8]. *)

Inductive ev : nat -> Prop :=
| ev_0 : ev 0
| ev_SS : forall n : nat, ev n -> ev (S (S n)).

Definition ev_8' : ev 8 :=
  (* FILL IN HERE *) admit.

(** **** Exercise: 2 stars, [conj_fact]  *)
(** Construct a proof object demonstrating the following proposition. *)

Definition conj_fact : forall P Q R, P /\ Q -> Q /\ R -> P /\ R :=
  (* FILL IN HERE *) admit.

(** **** Exercise: 2 stars, (or_commut)  *)
(** Write down an explicit proof object for [or_commut]. *)

Definition or_commut : forall P Q, P \/ Q -> Q \/ P :=
  (* FILL IN HERE *) admit.

(* ============== Abstract Data Types ============== *)

Require Import
        ADT.ADT
        ADT.ADTSigNotations
        ADT.BisimilarADTs
        ADT.ADTNotations.

Open Scope string_scope.
Open Scope ADTParsing_scope.
Open Scope ADTSig_scope.

(* Our ADT definitions use string for operation names; it's convenient to *)
(* declare aliases for these strings. *)
Definition new := "new".
Definition count := "count".
Definition add := "add".
Definition isempty := "isempty".

(* Here is the interface for our multiset ADT. One of the consequeneces of the *)
(* definition of ADT we're using is that every method needs to return an *)
(* instance of the rep type. Observer methods like count can just leave it unchanged. *)
Definition MultiSetT  :=
  ADTsignature { Constructor new : rep ,
                 Method add : rep * nat -> rep ,
                 Method isempty : rep -> rep * bool,
                 Method count : rep * nat -> rep * nat
               }.

(* Here is the list implementation of the multiset we saw in class. *)
Definition ListSet : ADT MultiSetT :=
  Def ADT { rep := list nat,
            Def Constructor0 new : rep := nil,
            Def Method1 add (r : rep) (n : nat) : rep := n :: r,
            Def Method0 isempty (r : rep) : rep * bool :=
              (r, match r with nil => true | _ => false end),
            Def Method1 count (r : rep) (n : nat) : rep * nat :=
              (r, fold_left (fun (count : nat) (m : nat) => if PeanoNat.Nat.eqb m n then S count else count) r 0)
      }.

(* And here's an implementation using a function. *)
Definition FunSet : ADT MultiSetT :=
  Def ADT { rep := (nat -> nat) * bool,
        Def Constructor0 new : rep := let new := fun m => 0 in (new, true),
        Def Method1 add (r : rep) (n : nat) : rep := (let add : nat -> nat := fun m => if PeanoNat.Nat.eqb m n then S (fst r m) else (fst r m) in add, false),
        Def Method0 isempty (r : rep) : rep * bool := (r, snd r),
        Def Method1 count (r : rep) (n : nat) : rep * nat :=
              (r, fst r n)
      }.

(* Here's a client of our multiset ADT that builds a multiset with a zero in it. *)
Example SetClient (a : DecoratedADT MultiSetT) :=
  let (X, x) := a in
  snd (x-->count (x--> add (x-->add (x-->new) 0) 0) 0).

(* We can pass the client an implementation to pull out the result. *)
Example SetClient_ex_1 : SetClient FunSet = 2.
Proof. compute. reflexivity. Qed.

Example SetClient_ex_2 : SetClient ListSet = 2.
Proof. compute. reflexivity. Qed.

(* We can't define a client that leaks the representation type of the ADT. *)
Fail Definition SetClient_1 (a : DecoratedADT MultiSetT) :=
  let (X, x) := a in x-->new.


(** **** Exercise: 2 stars *)
(** Here's an interface for a queue with enqueue and dequeue operations; *)
(* write an implementation for this interface using a list. *)

Definition enqueue := "enqueue".
Definition dequeue := "dequeue".

Definition StackT  :=
  ADTsignature { Constructor new : rep ,
                 Method enqueue : rep * nat -> rep ,
                 Method dequeue : rep -> rep * nat
               }.

Definition ListStack : ADT StackT := admit. (* Fill In Here *)

(** **** Exercise: 1 stars *)
(** Make sure that your stack implementation behaves as expected *)

Example ListStack_1 :
  let (X, x) := ListStack in
  snd (x-->dequeue (x-->enqueue (x-->enqueue (x-->enqueue (x-->new) 0) 1) 2)) = 0.
Proof. compute. Admitted.

Example ListStack_2 :
  let (X, x) := ListStack in
  snd (x-->dequeue (fst (x-->dequeue (x-->enqueue (x-->enqueue (x-->enqueue (x-->new) 0) 1) 2)))) = 1.
Proof. compute. Admitted.

(* We can prove that ListSet and FunSet are equivalent (or bisimilar) by *)
(* defining a relation between their representation types, and showing *)
(* that the methods preserve this relation. This guarantees that these *)
(* two implementations are indistinguishable from a client perspective. *)

Inductive ListSet_FunSet_R : list nat -> ((nat -> nat) * bool) -> Prop :=
| nil_R : ListSet_FunSet_R [ ] (fun _ : nat => 0, true)
| cons_R : forall n l' f' f b,
    ListSet_FunSet_R l' (f', b)
    -> f n = S (f' n)
    -> (forall m, m <> n -> f m = f' m)
    -> ListSet_FunSet_R (n :: l') (f, false).

Lemma FunSet_ListSet_Equivalent
  : BisimiliarADT ListSet FunSet ListSet_FunSet_R.
Proof.
  (* We start off by using a handy tactic that creates cases for each method. *)
  start_bisimilarity_proof; intros.
  (* Now we just need to show that each method respects the Relation we defined above. *)
  - econstructor.
  - destruct r_n as [f' b].
    econstructor; eauto.
    rewrite <- EqNat.beq_nat_refl; reflexivity.
    intros; apply PeanoNat.Nat.eqb_neq in H0; rewrite H0.
    reflexivity.
  - split; eauto.
    inversion H; subst; eauto.
  - split; eauto.
    rewrite (plus_n_O).
    generalize 0.
    induction H; simpl.
    + reflexivity.
    + simpl in *; intros.
      destruct (PeanoNat.Nat.eqb n d) eqn: ?.
      * apply EqNat.beq_nat_true in Heqb0; subst.
        rewrite H0, IHListSet_FunSet_R; Omega.omega.
      * apply EqNat.beq_nat_false in Heqb0; subst.
        rewrite H1, IHListSet_FunSet_R; Omega.omega.
Qed.

Lemma SetClient_Representation_Independence
      rep1 new1 add1 isempty1 count1
      rep2 new2 add2 isempty2 count2
    (Impl1 := Def ADT { rep := rep1,
            Def Constructor0 new : rep := new1,
            Def Method1 add (r : rep) (n : nat) : rep := add1 r n,
            Def Method0 isempty (r : rep) : rep * bool := isempty1 r,
            Def Method1 count (r : rep) (n : nat) : rep * nat := count1 r n
      })
    (Impl2 := Def ADT { rep := rep2,
            Def Constructor0 new : rep := new2,
            Def Method1 add (r : rep) (n : nat) : rep := add2 r n,
            Def Method0 isempty (r : rep) : rep * bool := isempty2 r,
            Def Method1 count (r : rep) (n : nat) : rep * nat := count2 r n
      })
    (AbsR : rep1 -> rep2 -> Prop)
    : BisimiliarADT Impl1 Impl2 AbsR
    -> SetClient Impl1 = SetClient Impl2.
Proof.
  intros.
  apply proveBisimilarADT in H;
    destruct H as [new_sim [add_sim [isempty_sim [count_sim _] ] ] ]; simpl in *.
  apply count_sim.
  apply add_sim.
  apply add_sim.
  apply new_sim.
Qed.

(* Here is the "bad" list implementation of the multiset we saw in class. *)
Definition ListSet_bad : ADT MultiSetT :=
  Def ADT { rep := list nat,
            Def Constructor0 new : rep := nil,
            Def Method1 add (r : rep) (n : nat) : rep := n :: r,
            Def Method0 isempty (r : rep) : rep * bool :=
              (r, match r with nil => true | _ => false end),
            Def Method1 count (r : rep) (n : nat) : rep * nat :=
              (r, 0)
      }.

(* We can show that it is impossible for this bad list implementation to be *)
(* swapped in for the original one. *)

Lemma ListSet_ListSet_bad_not_equiv
  : ~ exists R,
        BisimiliarADT ListSet ListSet_bad R.
Proof.
  intros [R Sim_R].
  generalize (SetClient_Representation_Independence _ _ _ _ _ _ _ _ _ _ R Sim_R).
  simpl.
  intro.
  discriminate.
Qed.

(* Here is yet another implementation of multisets using lists. *)

Definition rev_ListSet : ADT MultiSetT :=
  Def ADT { rep := list nat,
            Def Constructor0 new : rep := nil,
            Def Method1 add (r : rep) (n : nat) : rep := r ++ (cons n nil),

            Def Method0 isempty (r : rep) : rep * bool :=
              (r, match r with nil => true | _ => false end),
            Def Method1 count (r : rep) (n : nat) : rep * nat :=
                (r,
                 fold_left (fun (count : nat) (m : nat) => if PeanoNat.Nat.eqb m n then S count else count) r 0)
      }.

(** **** Exercise: 3 stars *)
(** Prove that this implementation is equivalent to ListSet by  *)
(* defining a relation between lists and completing the proof of  *)
(* ListSet_revList_Set_equiv. You may find count_succ useful. *)

Lemma count_succ
  : forall l n d,
    S (fold_left (fun count0 m : nat => if PeanoNat.Nat.eqb m d then S count0 else count0) l n) =
    fold_left (fun count0 m : nat => if PeanoNat.Nat.eqb m d then S count0 else count0) l (S n).
Proof.
  intros.
  rewrite <- !fold_left_rev_right;
    induction (rev l); simpl; eauto.
  destruct (PeanoNat.Nat.eqb a d);
    rewrite IHl0; eauto.
Qed.


Definition ListSet_rev_R (l1 l2: list nat) : Prop := admit.

Lemma FunSet_rev_Equivalent
  : BisimiliarADT ListSet rev_ListSet ListSet_rev_R.
Proof.
  (* We start off by using a handy tactic that creates cases for each method. *)
  start_bisimilarity_proof; intros.
  - (* new method *) admit.
  - (* add method *) admit.
  - (* isempty method *) admit.
  - (* count method *) admit.
Admitted.
